@@ -50,14 +50,17 @@ def get_first_visible_element(context, elements):
     for element in elements:
         context.klity.trace(f" ID: {element.get_attribute('id')}")
         context.klity.trace(f" displayed: {element.is_displayed()}")
-        context.klity.trace(f" invisible: {element.get_attribute('class').lower()}")
-        if (
-            element.is_displayed()
-            and "invisible" not in element.get_attribute("class").lower()
+        context.klity.trace(f" class: {element.get_attribute('class').lower()}")
+        if element.is_displayed() and not any(
+            [
+                class_name in element.get_attribute("class").lower()
+                for class_name in context.klity.configuration["options"][
+                    "hidden_classes"
+                ]
+            ]
         ):
             return element
     context.klity.trace(f"=> No visible element")
-    raise TestException(context)
 
 
 def get_value(context, value):
@@ -68,10 +71,10 @@ def get_value(context, value):
     variables = list(set(re.findall(r"(?<!\\)\$([^$]+)\$", value)))
     context.klity.trace(f"variables: {variables}")
     context.klity.trace(f"known variables: {context.klity.variables}")
-    context.klity.trace(f"=> value: {value}")
+    context.klity.trace(f"before => value: {value}")
     for variable in variables:
         value = value.replace(f"${variable}$", context.klity.variables[variable])
-    context.klity.trace(f"=> value: {value}")
+    context.klity.trace(f"after => value: {value}")
 
     # Replacing constants
     constants = list(set(re.findall(r"(?<!\\)#([^#]+)#", value)))
@@ -114,6 +117,8 @@ def get_value(context, value):
     # Replacing escaped characters
     value = value.replace("\$", "$")
     value = value.replace("\#", "#")
+    value = value.replace("\\n", "\n")
+    context.klity.trace(f"=> final value: {value}")
     return value
 
 
@@ -127,7 +132,7 @@ class Finder:
     find: FunctionType
 
 
-def find(context, name, finders, visible, value, field=None):
+def find(context, name, finders, visible, value=None, field=None):
     elements = []
     for finder in finders:
         context.klity.trace(f"  {finder.description}")
@@ -139,22 +144,19 @@ def find(context, name, finders, visible, value, field=None):
         except InvalidSelectorException:
             elements = []
         if len(elements) > 0:
-            break
+            context.klity.trace(f"=> {len(elements)} elements found.")
+            if visible:
+                # Looking for a visible element only
+                context.klity.trace("  Looking for a visible element.")
+                element = get_first_visible_element(context, elements)
+                if element is not None:
+                    return element
+            else:
+                return elements[0]
     if len(elements) == 0:
         context.klity.trace(f"=> {name} '{value}' not found.")
         raise TestException(context)
 
-    context.klity.trace(f"=> {len(elements)} elements found.")
-    if visible:
-        # Looking for a visible element only
-        context.klity.trace("  Looking for a visible element.")
-        element = get_first_visible_element(context, elements)
-        if element is None:
-            raise TestException(context)
-        else:
-            return element
-    return elements[0]
-
 
 def find_button(context, value, visible=True):
     context.klity.trace(f"Looking for button: '{value}'")
@@ -225,6 +227,12 @@ def find_button(context, value, visible=True):
                 f"//*[self::a and contains(.,{normalize_for_xpath(x)})]"
             ),
         ),
+        Finder(
+            f"//img[@alt={normalize_for_xpath(value)}]",
+            lambda v: context.browser.find_elements_by_xpath(
+                f"//img[@alt={normalize_for_xpath(v)}]"
+            )
+        ),
     ]
     return find(context, "Button", finders, visible, value)
 
@@ -533,11 +541,11 @@ def find_element(context, value, visible=True):
     context.klity.trace(f"Looking for element: '{value}'")
     finders = [
         Finder(
-            "//*[not(self::p) and not(self::div) and not(self::li)"
-            f" and normalize-space()={normalize_for_xpath(value)}]",
+            "//*[not(self::p) and not(self::div) and not(self::li) and"
+                f" contains(normalize-space(text()), {normalize_for_xpath(value)})]",
             lambda v: context.browser.find_elements_by_xpath(
-                "//*[not(self::p) and not(self::div) and not(self::li)"
-                f" and normalize-space()={normalize_for_xpath(v)}]",
+                "//*[not(self::p) and not(self::div) and not(self::li) and"
+                f" contains(normalize-space(text()), {normalize_for_xpath(v)})]",
             ),
         ),
         Finder(
@@ -565,9 +573,9 @@ def find_element(context, value, visible=True):
             )
         ),
         Finder(
-            f"//*[normalize-space()={normalize_for_css(value)}]",
+            f"//*[contains(normalize-space(text()), {normalize_for_xpath(value)})]",
             lambda v: context.browser.find_elements_by_xpath(
-                f"//*[normalize-space()={normalize_for_css(v)}]",
+                f"//*[contains(normalize-space(text()), {normalize_for_xpath(v)})]",
             ),
         ),
     ]
@@ -586,36 +594,43 @@ def find_input(context, value, visible=True):
             lambda v: context.browser.find_elements_by_name(v),
         ),
         Finder(
-            f"//*[self::input and normalize-space(.)={normalize_for_xpath(value)}]",
+            "//*[(self::input or self::textarea) and "
+            f"normalize-space(.)={normalize_for_xpath(value)}]",
             lambda v: context.browser.find_elements_by_xpath(
-                f"//*[self::input and normalize-space(.)={normalize_for_xpath(v)}]"
+                "//*[(self::input or self::textarea) and "
+                f"normalize-space(.)={normalize_for_xpath(v)}]"
             ),
         ),
         Finder(
-            f"//*[self::input and contains(.,{normalize_for_xpath(value)})]",
+            "//*[(self::input or self::textarea) and "
+            f"contains(.,{normalize_for_xpath(value)})]",
             lambda v: context.browser.find_elements_by_xpath(
-                f"//*[self::input and contains(.,{normalize_for_xpath(v)})]"
+                "//*[(self::input or self::textarea) and "
+                f"contains(.,{normalize_for_xpath(v)})]"
             ),
         ),
         Finder(
-            "//*[self::input and "
+            "//*[(self::input or self::textarea) and "
             f"normalize-space(@value)={normalize_for_xpath(value)}]",
             lambda v: context.browser.find_elements_by_xpath(
-                f"//*[self::input and normalize-space(@value)={normalize_for_xpath(v)}]"
+                "//*[(self::input or self::textarea) and "
+                f"normalize-space(@value)={normalize_for_xpath(v)}]"
             ),
         ),
         Finder(
-            "//*[self::input and "
+            "//*[(self::input or self::textarea) and "
             f"normalize-space(@title)={normalize_for_xpath(value)}]",
             lambda v: context.browser.find_elements_by_xpath(
-                f"//*[self::input and normalize-space(@title)={normalize_for_xpath(v)}]"
+                "//*[(self::input or self::textarea) and "
+                f"normalize-space(@title)={normalize_for_xpath(v)}]"
             ),
         ),
         Finder(
-            "//*[self::input and "
+            "//*[(self::input or self::textarea) and "
             f"normalize-space(@placeholder)={normalize_for_xpath(value)}]",
             lambda v: context.browser.find_elements_by_xpath(
-                f"//*[self::input and normalize-space(@placeholder)={normalize_for_xpath(v)}]"
+                "//*[(self::input or self::textarea) and "
+                f"normalize-space(@placeholder)={normalize_for_xpath(v)}]"
             ),
         ),
     ]
@@ -639,27 +654,43 @@ def find_checkbox(context, value, visible=True):
     context.klity.trace(f"Looking for field: '{value}'")
     finders = [
         Finder(
-            f"input[type='checkbox'][value={normalize_for_xpath(value)}]",
-            lambda v: context.browser.find_elements_by_css_selector(
-                f"input[type='checkbox'][value={normalize_for_xpath(v)}]"
+            "//input[@type='checkbox']"
+            f"[normalize-space(@value)={normalize_for_xpath(value)}]",
+            lambda v: context.browser.find_elements_by_xpath(
+                "//input[@type='checkbox']"
+                f"[normalize-space(@value)={normalize_for_xpath(v)}]"
             ),
         ),
         Finder(
-            f"input[type='checkbox'][id={normalize_for_xpath(value)}]",
-            lambda v: context.browser.find_elements_by_css_selector(
-                f"input[type='checkbox'][id={normalize_for_xpath(v)}]"
+            "//input[@type='checkbox']"
+            f"[normalize-space(@id)={normalize_for_xpath(value)}]",
+            lambda v: context.browser.find_elements_by_xpath(
+                "//input[@type='checkbox']"
+                f"[normalize-space(@id)={normalize_for_xpath(v)}]"
             ),
         ),
         Finder(
-            f"input[type='checkbox'][name={normalize_for_xpath(value)}]",
-            lambda v: context.browser.find_elements_by_css_selector(
-                f"input[type='checkbox'][name={normalize_for_xpath(v)}]"
+            "//input[@type='checkbox']"
+            f"[normalize-space(@name)={normalize_for_xpath(value)}]",
+            lambda v: context.browser.find_elements_by_xpath(
+                "//input[@type='checkbox']"
+                f"[normalize-space(@name)={normalize_for_xpath(v)}]"
             ),
         ),
         Finder(
-            f"input[type='checkbox'][title={normalize_for_xpath(value)}]",
-            lambda v: context.browser.find_elements_by_css_selector(
-                f"input[type='checkbox'][title={normalize_for_xpath(v)}]"
+            "//input[@type='checkbox']"
+            f"[normalize-space(@title)={normalize_for_xpath(value)}]",
+            lambda v: context.browser.find_elements_by_xpath(
+                "//input[@type='checkbox']"
+                f"[normalize-space(@title)={normalize_for_xpath(v)}]"
+            ),
+        ),
+        Finder(
+            "//input[@type='checkbox']/parent::*"
+            f"[normalize-space(.)={normalize_for_xpath(value)}]",
+            lambda v: context.browser.find_elements_by_xpath(
+                "//input[@type='checkbox']/parent::*"
+                f"[normalize-space(.)={normalize_for_xpath(v)}]"
             ),
         ),
     ]
@@ -679,6 +710,19 @@ def find_checkbox(context, value, visible=True):
     return find(context, "Checkbox", finders, visible, value)
 
 
+def find_submit(context, value=None):
+    context.klity.trace(f"Looking for submit input")
+    finders = [
+        Finder(
+            "//input[@type='submit']",
+            lambda v: context.browser.find_elements_by_xpath(
+                "//input[@type='submit']"
+            ),
+        ),
+    ]
+    return find(context, "Submit", finders, True)
+
+
 def find_recaptcha(context, value=None):
     context.klity.trace(f"Looking for ReCaptcha")
     finders = [
@@ -692,6 +736,43 @@ def find_recaptcha(context, value=None):
     return find(context, "ReCaptcha", finders, True, value)
 
 
+def find_image(context, value, visible=True):
+    context.klity.trace(f"Looking for image: '{value}'")
+    finders = [
+        Finder(
+            f"//img[@id={normalize_for_xpath(value)}]",
+            lambda v: context.browser.find_elements_by_xpath(
+                f"//img[@id={normalize_for_xpath(v)}]"
+            )
+        ),
+        Finder(
+            f"//img[@name={normalize_for_xpath(value)}]",
+            lambda v: context.browser.find_elements_by_xpath(
+                f"//img[@name={normalize_for_xpath(v)}]"
+            )
+        ),
+        Finder(
+            f"//img[@title={normalize_for_xpath(value)}]",
+            lambda v: context.browser.find_elements_by_xpath(
+                f"//img[@title={normalize_for_xpath(v)}]"
+            )
+        ),
+        Finder(
+            f"//img[@alt={normalize_for_xpath(value)}]",
+            lambda v: context.browser.find_elements_by_xpath(
+                f"//img[@alt={normalize_for_xpath(v)}]"
+            )
+        ),
+        Finder(
+            f"//img[contains(@src,{normalize_for_xpath(value)})]",
+            lambda v: context.browser.find_elements_by_xpath(
+                f"//img[contains(@src,{normalize_for_xpath(v)})]"
+            ),
+        ),
+    ]
+    return find(context, "Image", finders, visible, value)
+
+
 def find_file(name, path):
     for root, dirs, files in os.walk(path):
         if name in files:
@@ -746,14 +827,19 @@ def get_table(context):
     return array
 
 
-# TODO: Thoses two functions are not used for now...
+def type_text(context, text):
+    actions = ActionChains(context.browser)
+    for character in text:
+        context.klity.trace(f"Character {character}")
+        if character == "\n":
+            actions.send_keys(Keys.RETURN)
+        else:
+            actions.send_keys(character)
+    actions.perform()
+
+
+# TODO: This function is not used for now...
 def paste_from_clipboard(context):
     actions = ActionChains(context.browser)
     actions.send_keys(Keys.SHIFT, Keys.INSERT)
     actions.perform()
-
-
-def send_return(context):
-    actions = ActionChains(context.browser)
-    actions.send_keys(Keys.RETURN)
-    actions.perform()
